#include <iostream>
#include <fstream>
#include <sstream>
#include <signal.h>
#include <chrono>
#include <ctime>
#include <list>
#include <sys/stat.h>

#include "beegfs/beegfs_file_event_log.hpp"

/* The client needs a list of events to be logged in the config file
 * (beegfs-client.conf):
 * sysFileEventLogMask = flush,trunc,setattr,close,link-op
 * (Or any combination of the above)
 *
 * to deal with zfs issue, only configure close event:
 * sysFileEventLogMask = close
 */

/**
 * @brief Trivial JSON writer
 */
class JsonObject {

   public:

      template<typename T>
      JsonObject& keyValue(const std::string& key, const T& value) {
         if(!isFirstItem) {
            stream << ", ";

         } else {
            isFirstItem = false;
         }

         printValue(key);
         stream << ": ";
         printValue(value);
         return *this;
      }

      operator std::string() const {
         return str();
      }

      std::string str() const
      {
         return "{ " + stream.str() + " }";
      }
   protected:

      std::stringstream stream;
      bool isFirstItem=true;

      template<typename T>
      void printValue(const T& value) {
         stream << value;
      }

      void printValue(const std::string& value) {
         stream << "\"";
         writeEscaped(stream, value);
         stream << "\"";
      }

      void printValue(const char* value) {
         printValue(std::string(value));
      }

      void printValue(const JsonObject& json) {
         stream << json.str();
      }

      void writeEscaped(std::ostream& stream, const std::string& s)
      {
         for(const auto& x: s)
         {
            switch (x) {
               case 0x08:
                  stream << "\\b";
                  break;
               case 0x0c:
                  stream << "\\f";
                  break;
               case '\n':
                  stream << "\\n";
                  break;
               case '\\':
                  stream << "\\\\";
                  break;
               case '\t':
                  stream << "\\t";
                  break;
               case '\r':
                  stream << "\\r";
                  break;
               case '\"':
                  stream << "\\\"";
                  break;
               case '/':
                  stream << "\\/";
                  break;
               default:
                  stream << x;
            }
         }
      }

};


/**
 * Check if path exists
 * @return true only if the path exists, false else
 */
bool pathExists(const std::string& file) 
{
    struct stat buf;
    return (stat(file.c_str(), &buf) == 0);
}

/**
 * @brief Get current date/time
 * @return string formatted as YYYY-MM-DD.HH:mm:ss
 */
const std::string currentDateTime() 
{
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    // Visit http://en.cppreference.com/w/cpp/chrono/c/strftime
    // for more information about date/time format
    strftime(buf, sizeof(buf), "%Y-%m-%d-%X", &tstruct);

    return buf;
}

/**
 * @brief returns only the file path from a packet and makes sure, that leading '/' is stripped
 */
std::string renderPacket(const BeeGFS::packet& p)
{
   if (p.path.size() > 0 && p.path[0] == '/')
     return &p.path[1];
   else
     return p.path;
}

/**
 * @brief simple << operator that only prints the file path
 */
std::ostream& operator<<(std::ostream& os, const BeeGFS::packet& p)
{
   os << currentDateTime() << " " << renderPacket(p);
   return os;
}

void shutdown(int)
{
  exit(0);
}

/**
 * @brief dumps the complete buffer into a file and clears the buffer
 */
void FlushBufferToFile(std::list<std::string>& buffer, const std::string& destDir, const std::string& fname)
{
  if (!buffer.empty())
  {
    std::string destination = destDir;
    destination = destination+"/"+currentDateTime()+fname; // ToDo: use proper path function
    std::ofstream writer(destination, std::ios::out);

    for (auto const& i : buffer)
    {
       writer << i << std::endl;
    }
    buffer.clear();
  }
}

/*
* this program implements the following approach: 
* - there is a time window (e.g. 10 seconds) that defines the time period for which events shall be received and buffered in memory
* - this approach has been chosen so that the filesystem is not overloaded when many events are received
* - once the time window has passed, the events are stored into a file
* - hence, that file contains only events for a period that is defined by the time window (e.g. 10 seconds)
* - only the relative file path for the event is stored in the file
* - the loop below makes sure, that a file is written and the buffer is flushed at the appropriate point in time
* - all of that is single threaded, just to keep things simple
*/
int main(const int argc, const char** argv)
{
   if (argc < 4) {
      std::cout << "BeeGFS File Event Recorder\n"
                   "Usage:\n"
                   "  beegfs-event-recorder <socket> <destination-dir> <collection-window> [verbose]\n\n"
                   "  parameter <socket>:\n"
                   "  The medatada server has to be pointed to the socket, so that it knows where to\n"
                   "  send the event log. Set\n"
                   "     sysFileEventLogTarget = unix://<path>\n"
                   "  in beegfs-meta.conf.\n"
                   "  The clients need a list of events to be logged in the config file (beegfs-client.conf):\n"
                   "    sysFileEventLogMask = flush,trunc,setattr,close,link-op\n"
                   "  (Or any combination of the above)\n\n"
                   "  parameter <destination-dir>:\n"
                   "  destination directory, where recorder events shall be stored.\n"
                   "  the file names will be chosen using the pattern 'YYYY-mm-dd-HH-MM-ss-events.log'\n\n"
                   "  parameter <collection-window>:\n"
                   "  number in seconds for which events should be recorder into one file.\n"
                   "  if no events are received within the collection window, then no file will be written\n\n"
                   "  optional parameter verbose:\n"
                   "  prints out some debugging output. This should not be set druign productive usage.\n"
                   "  Example:\n"
                   "     Enter the following line in beegfs-meta.conf:\n"
                   "       sysFileEventLogTarget = unix:///var/beegfslog\n"
                   "     and run \n"
                   "     beegfs-event-recorder /var/beegfslog /tmp/events 10\n"
                << std::endl;
      return EXIT_FAILURE;
   }

    signal(SIGINT,  shutdown);

    BeeGFS::FileEventReceiver receiver(argv[1]);

    std::cout << JsonObject().keyValue("EventListener",
                           JsonObject().keyValue("Socket", argv[1])
                                       .keyValue("VersionMajor", BEEGFS_EVENTLOG_FORMAT_MAJOR)
                                       .keyValue("VersionMinor", BEEGFS_EVENTLOG_FORMAT_MINOR)
                  ).str() << std::endl;

    const std::string destDir = argv[2];
    const std::string outFile = "-events.log";

    int windowSecs = 10;

    try
    {
       windowSecs = std::stoi(argv[3]);
    }
    catch (...)
    {
       std::cerr << "parameter collection-window must be an integer. Parsing failed for '" << argv[3] << "'." << std::endl;
       exit(1);
    }

    const int delta = 1;
    bool verbose = false;

    if (argc > 4 && strcmp(argv[4], "verbose") == 0)
        verbose = true;

    // check directory
    if (!pathExists(destDir))
    {
       std::cerr << "directory " << destDir << " does not exist" << std::endl;
       exit(1);
    }


    std::list<std::string> buffer;

    std::chrono::time_point<std::chrono::system_clock> lastTime = std::chrono::system_clock::now();

    while (true)
    {
       // ToDo: exception handling ?
       using BeeGFS::FileEventReceiver;

       if (verbose)
           std::cout << "waiting for event..." << std::endl;

       const auto data = receiver.read();
       switch (data.first) {
          case FileEventReceiver::ReadErrorCode::Success:
          {
             if (verbose)
                 std::cout << "pushing event to buffer..." << std::endl;
             
             // insert the event to the buffer
             buffer.push_back(renderPacket(data.second));

             std::chrono::time_point<std::chrono::system_clock> currentTime = std::chrono::system_clock::now();
             auto seconds = std::chrono::duration_cast<std::chrono::seconds>(currentTime - lastTime);

             // seconds.count() >= windowSecs: this covers the case, when an event has been received, but now the window already passed
             // !receiver.hasData(windowSecs -seconds.count() + delta: 
             //    this will block until the current window has passed
             //    and it will return true, if there is no pending event in the socket queue
             //    otherwise, the loop shall just continue and process the pending events
             if ((seconds.count() >= windowSecs) || (!receiver.hasData(windowSecs - seconds.count() + delta)))
             {
               if (verbose)
                   std::cout << "writing to file YYYY-mm-dd-HH:MM:ss" << outFile << " because since " << windowSecs << " no file was written." << std::endl;
               FlushBufferToFile(buffer, destDir, outFile); // flush buffer to file and clear buffer
               lastTime = std::chrono::system_clock::now(); // remember the last time when the buffer was written
             }
             // else we are still within one window...
          }
          break;
          case FileEventReceiver::ReadErrorCode::VersionMismatch:
             std::cerr << "Invalid Packet Version" << std::endl;
          break;
          case FileEventReceiver::ReadErrorCode::InvalidSize:
             std::cerr << "Invalid Packet Size" << std::endl;
          break;
          case FileEventReceiver::ReadErrorCode::ReadFailed:
             std::cerr << "Read Failed" << std::endl;
          break;
       }
    }

    return 0;
}
