#!/bin/sh

set -e

now=$(date +%s)

### DEFAULT CONFIGURATION
# BeeGFS mount point
beegfs_mnt=/mnt/beegfs
# Directory containing the events files
event_files_dir=events
# Only event files older than interval seconds are processed
interval=120
# Run whenever files are written to events dir
use_inotify=false
# Print which files are processed but don't actually touch them
dry_run=false
# Print verbose info
verbose=false

### HELPER FUNCTIONS
print_verbose() {
    if $verbose ; then
        echo "$1" >&2
    fi
}

# Parse time stamps to unix epoch.
parse_ts() {
    IFS=-;
    read -r year month day t <<EOF
$ts
EOF
    date -d "$year-$month-$day $t" "+%s"
}

# Generate time string for filename from epoch
unparse_ts() {
    date -d "@$1" +%F-%T
}

usage() {
    echo >&2 "Usage: $0 [-e event_files_dir] [-t interval] [-p beegfs_mnt]"
    echo >&2 "                      [-i use_inotify] [-n dry_run] [-v verbose]"
    echo >&2 ""
    echo >&2 "Paramters:"
    echo >&2 "  -e: Directory containing the events files (default \"events\")"
    echo >&2 "  -t: Only event files older than interval seconds are processed"
    echo >&2 "      (default 120)"
    echo >&2 "  -p: BeeGFS mount point (default \"/mnt/beegfs\")"
    echo >&2 "  -i: Run whenever files are written to events dir"
    echo >&2 "      (default false, polls every 10 seconds)"
    echo >&2 "  -n: Print which files are processed but don't actually touch them"
    echo >&2 "      (default false)"
    echo >&2 "  -v: Print verbose info"
    echo >&2 "      (default false)"
}

### CORE FUNCTIONS
# Read event filenames and return list of epochs
read_event_files() {
    for file in "$event_files_dir"/*-events.log ; do
        # no nullglob in POSIX sh
        if ! [ -f "$file" ] ; then
            continue
        fi
        file=$(basename "$file")
        ts=${file%-events.log}
        parse_ts
    done
}

# Only return epochs older than now - interval
filter_event_files() {
    while [ $# -gt 0 ] ; do
        epoch=$1
        if [ $((now-epoch-interval)) -ge 0 ] ; then
            echo "$epoch"
        fi
        shift
    done
}

# Check if event file exists and process it
process_event_file() {
    event_fpath="$event_files_dir/$1"
    if ! [ -f "$event_fpath" ] ; then
        print_verbose "  Couldn't open event file $file"
        return 0
    fi
    while read -r file ; do
        fpath="$beegfs_mnt/$file"
        if [ -f "$fpath" ] ; then
            print_verbose "  Processing $fpath"
            if ! $dry_run ; then
                touch -r "$fpath" "$fpath" || echo "Error: touch failed on file $fpath"
            fi
        else
            print_verbose "  Warning: File \"$fpath\" not found. Skipping."
            continue
        fi
    done <"$event_fpath"
    if ! $dry_run ; then
        rm "$event_fpath"
    fi
}

# Main function
run() {
    now=$(date +%s)
    epochs=$(filter_event_files $(read_event_files))
    for epoch in $epochs ; do
        fname="$(unparse_ts "$epoch")-events.log"
        print_verbose "Processing event file $fname"
        process_event_file "$fname"
    done
}

### PARSE ARGUMENTS
while getopts e:t:p:inv o ; do
    case "$o" in
    e)
        event_files_dir="$OPTARG"
        if ! [ -d "$event_files_dir" ] ; then
            echo "Event file dir \"$event_files_dir\" does not exist. Exiting."
            exit 1
        fi
        ;;
    t)
        interval="$OPTARG"
        if ! [ "$interval" -eq "$interval" ] >/dev/null 2>&1 ; then
            echo "Invalid interval \"$interval\". Exiting."
            exit 1
        fi
        ;;
    p)
        beegfs_mnt="$OPTARG"
        if ! [ -d "$beegfs_mnt" ] ; then
            echo "Mount point \"$beegfs_mnt\" does not exist. Exiting."
            exit 1
        fi
        ;;
    i)
        use_inotify=true
        ;;
    n)
        dry_run=true
        ;;
    v)
        verbose=true
        ;;
    [?])
        usage
        exit 1
        ;;
    esac
done


if $use_inotify ; then
    print_verbose "Watching directory \"$event_files_dir\" with inotify"
    while true ; do
        inotifywait -r -e close_write "$event_files_dir"
        run
    done
else
    while true ; do
        print_verbose "Polling directory \"$event_files_dir\" for changes"
        run
        sleep 10
    done
fi
